@@ -9,19 +9,19 @@ files.  At this time it has cache helpers and understands the
 if-modified-since request header.
 """
 
-import os, time, mimetypes
+import os, time, mimetypes, zipfile, tarfile
 from paste.httpexceptions import *
 from paste.httpheaders import *
 
 CACHE_SIZE = 4096
 BLOCK_SIZE = 4096 * 16
 
-__all__ = ['DataApp', 'FileApp']
+__all__ = ['DataApp', 'FileApp', 'DirectoryApp', 'ArchiveStore']
 
 class DataApp(object):
     """
     Returns an application that will send content in a single chunk,
-    this application has support for setting cashe-control and for
+    this application has support for setting cache-control and for
     responding to conditional (or HEAD) requests.
 
     Constructor Arguments:
@@ -53,12 +53,18 @@ class DataApp(object):
         like changing ``Last-Modified`` and ``Content-Length`` headers.
 
     """
-    def __init__(self, content, headers=None, **kwargs):
+
+    allowed_methods = ('GET', 'HEAD')
+
+    def __init__(self, content, headers=None, allowed_methods=None,
+                 **kwargs):
         assert isinstance(headers, (type(None), list))
         self.expires = None
         self.content = None
         self.content_length = None
         self.last_modified = 0
+        if allowed_methods is not None:
+            self.allowed_methods = allowed_methods
         self.headers = headers or []
         for (k, v) in kwargs.items():
             header = get_header(k)
@@ -73,9 +79,12 @@ class DataApp(object):
         self.expires = CACHE_CONTROL.apply(self.headers, **kwargs) or None
         return self
 
-    def set_content(self, content):
+    def set_content(self, content, last_modified=None):
         assert content is not None
-        self.last_modified = time.time()
+        if last_modified is None:
+            self.last_modified = time.time()
+        else:
+            self.last_modified = last_modified
         self.content = content
         self.content_length = len(content)
         LAST_MODIFIED.update(self.headers, time=self.last_modified)
@@ -86,23 +95,24 @@ class DataApp(object):
         return self
 
     def __call__(self, environ, start_response):
+        method = environ['REQUEST_METHOD'].upper()
+        if method not in self.allowed_methods:
+            exc = HTTPMethodNotAllowed(
+                'You cannot %s a file' % method,
+                headers=[('Allow', ','.join(self.allowed_methods))])
+            return exc(environ, start_response)
+        return self.get(environ, start_response)
+
+    def calculate_etag(self):
+        return '"%s-%s"' % (self.last_modified, self.content_length)
+
+    def get(self, environ, start_response):
         headers = self.headers[:]
-        current_etag = str(self.last_modified)
+        current_etag = self.calculate_etag()
         ETAG.update(headers, current_etag)
         if self.expires is not None:
             EXPIRES.update(headers, delta=self.expires)
 
-        try:
-            client_clock = IF_MODIFIED_SINCE.parse(environ)
-            if client_clock >= int(self.last_modified):
-                # horribly inefficient, n^2 performance, yuck!
-                for head in list_headers(entity=True):
-                    head.delete(headers)
-                start_response('304 Not Modified', headers)
-                return [''] # empty body
-        except HTTPBadRequest, exce:
-            return exce.wsgi_application(environ, start_response)
-
         try:
             client_etags = IF_NONE_MATCH.parse(environ)
             if client_etags:
@@ -116,6 +126,22 @@ class DataApp(object):
         except HTTPBadRequest, exce:
             return exce.wsgi_application(environ, start_response)
 
+        # If we get If-None-Match and If-Modified-Since, and
+        # If-None-Match doesn't match, then we should not try to
+        # figure out If-Modified-Since (which has 1-second granularity
+        # and just isn't as accurate)
+        if not client_etags:
+            try:
+                client_clock = IF_MODIFIED_SINCE.parse(environ)
+                if client_clock >= int(self.last_modified):
+                    # horribly inefficient, n^2 performance, yuck!
+                    for head in list_headers(entity=True):
+                        head.delete(headers)
+                    start_response('304 Not Modified', headers)
+                    return [''] # empty body
+            except HTTPBadRequest, exce:
+                return exce.wsgi_application(environ, start_response)
+
         (lower, upper) = (0, self.content_length - 1)
         range = RANGE.parse(environ)
         if range and 'bytes' == range[0] and 1 == len(range[1]):
@@ -149,13 +175,16 @@ class FileApp(DataApp):
 
     def __init__(self, filename, headers=None, **kwargs):
         self.filename = filename
-        content_type, content_encoding = mimetypes.guess_type(self.filename)
+        content_type, content_encoding = self.guess_type()
         if content_type and 'content_type' not in kwargs:
             kwargs['content_type'] = content_type
         if content_encoding and 'content_encoding' not in kwargs:
             kwargs['content_encoding'] = content_encoding
         DataApp.__init__(self, None, headers, **kwargs)
 
+    def guess_type(self):
+        return mimetypes.guess_type(self.filename)
+
     def update(self, force=False):
         stat = os.stat(self.filename)
         if not force and stat.st_mtime == self.last_modified:
@@ -163,7 +192,7 @@ class FileApp(DataApp):
         self.last_modified = stat.st_mtime
         if stat.st_size < CACHE_SIZE:
             fh = open(self.filename,"rb")
-            self.set_content(fh.read())
+            self.set_content(fh.read(), stat.st_mtime)
             fh.close()
         else:
             self.content = None
@@ -172,12 +201,18 @@ class FileApp(DataApp):
             # called
             LAST_MODIFIED.update(self.headers, time=self.last_modified)
 
-    def __call__(self, environ, start_response):
+    def get(self, environ, start_response):
+        is_head = environ['REQUEST_METHOD'].upper() == 'HEAD'
         if 'max-age=0' in CACHE_CONTROL(environ).lower():
             self.update(force=True) # RFC 2616 13.2.6
         else:
             self.update()
         if not self.content:
+            if not os.path.exists(self.filename):
+                exc = HTTPNotFound(
+                    'The resource does not exist',
+                    comment="No file at %r" % self.filename)
+                return exc(environ, start_response)
             try:
                 file = open(self.filename, 'rb')
             except (IOError, OSError), e:
@@ -185,15 +220,23 @@ class FileApp(DataApp):
                     'You are not permitted to view this file (%s)' % e)
                 return exc.wsgi_application(
                     environ, start_response)
-        retval = DataApp.__call__(self, environ, start_response)
+        retval = DataApp.get(self, environ, start_response)
         if isinstance(retval, list):
             # cached content, exception, or not-modified
+            if is_head:
+                return ['']
             return retval
         (lower, content_length) = retval
+        if is_head:
+            return ['']
         file.seek(lower)
-        return _FileIter(file, size=content_length)
+        file_wrapper = environ.get('wsgi.file_wrapper', None)
+        if file_wrapper:
+            return file_wrapper(file, BLOCK_SIZE)
+        else:
+            return _FileIter(file, size=content_length)
 
-class _FileIter:
+class _FileIter(object):
 
     def __init__(self, file, block_size=None, size=None):
         self.file = file
@@ -217,3 +260,90 @@ class _FileIter:
     def close(self):
         self.file.close()
 
+
+class DirectoryApp(object):
+    """
+    Returns an application that dispatches requests to corresponding FileApps based on PATH_INFO.
+    FileApp instances are cached. This app makes sure not to serve any files that are not in a subdirectory.
+    To customize FileApp creation override ``DirectoryApp.make_fileapp``
+    """
+
+    def __init__(self, path):
+        self.path = os.path.abspath(path)
+        if not self.path.endswith(os.path.sep):
+            self.path += os.path.sep
+        assert os.path.isdir(self.path)
+        self.cached_apps = {}
+
+    make_fileapp = FileApp
+
+    def __call__(self, environ, start_response):
+        path_info = environ['PATH_INFO']
+        app = self.cached_apps.get(path_info)
+        if app is None:
+            path = os.path.join(self.path, path_info.lstrip('/'))
+            if not os.path.normpath(path).startswith(self.path):
+                app = HTTPForbidden()
+            elif os.path.isfile(path):
+                app = self.make_fileapp(path)
+                self.cached_apps[path_info] = app
+            else:
+                app = HTTPNotFound(comment=path)
+        return app(environ, start_response)
+
+
+class ArchiveStore(object):
+    """
+    Returns an application that serves up a DataApp for items requested
+    in a given zip or tar archive.
+
+    Constructor Arguments:
+
+        ``filepath``    the path to the archive being served
+
+    ``cache_control()``
+
+        This method provides validated construction of the ``Cache-Control``
+        header as well as providing for automated filling out of the
+        ``EXPIRES`` header for HTTP/1.0 clients.
+    """
+
+    def __init__(self, filepath):
+        if zipfile.is_zipfile(filepath):
+            self.archive = zipfile.ZipFile(filepath,"r")
+        elif tarfile.is_tarfile(filepath):
+            self.archive = tarfile.TarFileCompat(filepath,"r")
+        else:
+            raise AssertionError("filepath '%s' is not a zip or tar " % filepath)
+        self.expires = None
+        self.last_modified = time.time()
+        self.cache = {}
+
+    def cache_control(self, **kwargs):
+        self.expires = CACHE_CONTROL.apply(self.headers, **kwargs) or None
+        return self
+
+    def __call__(self, environ, start_response):
+        path = environ.get("PATH_INFO","")
+        if path.startswith("/"):
+            path = path[1:]
+        application = self.cache.get(path)
+        if application:
+            return application(environ, start_response)
+        try:
+            info = self.archive.getinfo(path)
+        except KeyError:
+            exc = HTTPNotFound("The file requested, '%s', was not found." % path)
+            return exc.wsgi_application(environ, start_response)
+        if info.filename.endswith("/"):
+            exc = HTTPNotFound("Path requested, '%s', is not a file." % path)
+            return exc.wsgi_application(environ, start_response)
+        content_type, content_encoding = mimetypes.guess_type(info.filename)
+        app = DataApp(None, content_type = content_type,
+                            content_encoding = content_encoding)
+        app.set_content(self.archive.read(path),
+                time.mktime(info.date_time + (0,0,0)))
+        self.cache[path] = app
+        app.expires = self.expires
+        return app(environ, start_response)
+
