@@ -57,6 +57,10 @@ def html_unquote(s, encoding=None):
     u'\u1029'
     """
     if isinstance(s, str):
+        if s == '':
+            # workaround re.sub('', '', u'') returning '' < 2.5.2
+            # instead of u'' >= 2.5.2
+            return u''
         s = s.decode(encoding or default_encoding)
     return _unquote_re.sub(_entity_subber, s)
 
@@ -72,6 +76,13 @@ def no_quote(s):
     """
     return s
 
+_comment_quote_re = re.compile(r'\-\s*\>')
+def comment_quote(s):
+    """
+    Quote that makes sure text can't escape a comment
+    """
+    return _comment_quote_re.sub('-&gt', str(s))
+
 url_quote = urllib.quote
 url_unquote = urllib.unquote
 
